@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Copyright 2014 Tomo Krajina
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -21,22 +19,24 @@ import copy as mod_copy
 
 from . import utils as mod_utils
 
+from typing import *
+
 
 class GPXFieldTypeConverter:
-    def __init__(self, from_string, to_string):
+    def __init__(self, from_string: str, to_string: str) -> None:
         self.from_string = from_string
         self.to_string = to_string
 
 
 RE_TIMESTAMP = mod_re.compile(
     r'^([0-9]{4})-([0-9]{1,2})-([0-9]{1,2})[T ]([0-9]{1,2}):([0-9]{1,2}):([0-9]{1,2})'
-    r'(\.[0-9]{1,8})?(Z|[+-−][0-9]{2}:?(?:[0-9]{2})?)?$')
+    r'(\.[0-9]{1,15})?(Z|[+-−][0-9]{2}:?(?:[0-9]{2})?)?$')
 
 
 class SimpleTZ(mod_datetime.tzinfo):
     __slots__ = ('offset',)
 
-    def __init__(self, s=None):
+    def __init__(self, s: str="") -> None:
         self.offset = 0
         if s and len(s) >= 2:
             if s[0] in ('−', '-'):
@@ -53,25 +53,25 @@ class SimpleTZ(mod_datetime.tzinfo):
                 minute = 0
             self.offset = mult * (hour * 60 + minute)
 
-    def utcoffset(self, dt):
+    def utcoffset(self, dt: Optional[mod_datetime.datetime]) -> mod_datetime.timedelta:
         return mod_datetime.timedelta(minutes=self.offset)
 
-    def dst(self, dt):
+    def dst(self, dt: Optional[mod_datetime.datetime]) -> mod_datetime.timedelta:
         return mod_datetime.timedelta(0)
 
-    def tzname(self, dt):
+    def tzname(self, dt: Optional[mod_datetime.datetime]) -> str:
         if self.offset == 0:
             return 'Z'
         return '{:02}:{:02}'.format(self.offset // 60, self.offset % 60)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return 'SimpleTZ("{}")'.format(self.tzname(None))
 
-    def __eq__(self, other):
-        return self.offset == other.offset
+    def __eq__(self, other: Any) -> bool:
+        return self.offset == other.offset # type: ignore
 
 
-def parse_time(string):
+def parse_time(string: str) -> Optional[mod_datetime.datetime]:
     from . import gpx as mod_gpx
     if not string:
         return None
@@ -83,12 +83,12 @@ def parse_time(string):
             dt.append(int(f + "0" * (6 - len(f))))
         else:
             dt.append(0)
-        dt.append(SimpleTZ(m.group(8)))
-        return mod_datetime.datetime(*dt)
-    raise mod_gpx.GPXException('Invalid time: {0}'.format(string))
+        dt.append(SimpleTZ(m.group(8))) # type: ignore
+        return mod_datetime.datetime(*dt) # type: ignore
+    raise mod_gpx.GPXException(f'Invalid time: {string}')
 
 
-def format_time(time):
+def format_time(time: mod_datetime.datetime) -> str:
     offset = time.utcoffset()
     if not offset or offset == 0:
         tz = 'Z'
@@ -108,26 +108,28 @@ def format_time(time):
 
 
 class FloatConverter:
-    def __init__(self):
+    def __init__(self) -> None:
         self.from_string = lambda string : None if string is None else float(string.strip())
         self.to_string =   lambda flt    : mod_utils.make_str(flt)
 
 
 class IntConverter:
-    def __init__(self):
+    def __init__(self) -> None:
         self.from_string = lambda string: None if string is None else int(string.strip())
         self.to_string = lambda flt: str(flt)
 
 
 class TimeConverter:
-    def from_string(self, string):
+    def from_string(self, string: str) -> Optional[mod_datetime.datetime]:
         try:
             return parse_time(string)
         except:
             return None
 
-    def to_string(self, time):
-        return format_time(time) if time else None
+    def to_string(self, time: Optional[mod_datetime.datetime]) -> Optional[str]:
+        if time:
+            return format_time(time) if time else None
+        return None
 
 
 INT_TYPE = IntConverter()
@@ -141,15 +143,15 @@ TIME_TYPE = TimeConverter()
 
 
 class AbstractGPXField:
-    def __init__(self, attribute_field=None, is_list=None):
+    def __init__(self, attribute_field: Optional[str] = None, is_list: Optional[bool]=None) -> None:
         self.attribute_field = attribute_field
         self.is_list = is_list
-        self.attribute = False
+        self.attribute: Optional[str] = None
 
-    def from_xml(self, node, version):
+    def from_xml(self, node: str, version: str) -> Any:
         raise Exception('Not implemented')
 
-    def to_xml(self, value, version, nsmap):
+    def to_xml(self, value: Any, version: str, nsmap: Any) -> Optional[str]:
         raise Exception('Not implemented')
 
 
@@ -157,8 +159,8 @@ class GPXField(AbstractGPXField):
     """
     Used for to (de)serialize fields with simple field<->xml_tag mapping.
     """
-    def __init__(self, name, tag=None, attribute=None, type=None,
-                 possible=None, mandatory=None):
+    def __init__(self, name: Optional[str], tag: Optional[str]=None, attribute: Optional[str]=None, type: Any=None,
+                 possible: Optional[Iterable[str]]=None, mandatory: Optional[bool]=None) -> None:
         AbstractGPXField.__init__(self)
         self.name = name
         if tag and attribute:
@@ -166,7 +168,7 @@ class GPXField(AbstractGPXField):
             raise mod_gpx.GPXException('Only tag *or* attribute may be given!')
         if attribute:
             self.tag = None
-            self.attribute = name if attribute is True else attribute
+            self.attribute = attribute
         elif tag:
             self.tag = name if tag is True else tag
             self.attribute = None
@@ -177,7 +179,7 @@ class GPXField(AbstractGPXField):
         self.possible = possible
         self.mandatory = mandatory
 
-    def from_xml(self, node, version):
+    def from_xml(self, node: Any, version: str) -> Any:
         if self.attribute:
             if node is not None:
                 result = node.get(self.attribute)
@@ -190,7 +192,7 @@ class GPXField(AbstractGPXField):
         if result is None:
             if self.mandatory:
                 from . import gpx as mod_gpx
-                raise mod_gpx.GPXException('{0} is mandatory in {1} (got {2})'.format(self.name, self.tag, result))
+                raise mod_gpx.GPXException(f'{self.name} is mandatory in {self.tag} (got {result})')
             return None
 
         if self.type_converter:
@@ -198,36 +200,38 @@ class GPXField(AbstractGPXField):
                 result = self.type_converter.from_string(result)
             except Exception as e:
                 from . import gpx as mod_gpx
-                raise mod_gpx.GPXException('Invalid value for <{0}>... {1} ({2})'.format(self.tag, result, e))
+                raise mod_gpx.GPXException(f'Invalid value for <{self.tag}>... {result} ({e})')
 
         if self.possible:
             if not (result in self.possible):
                 from . import gpx as mod_gpx
-                raise mod_gpx.GPXException('Invalid value "{0}", possible: {1}'.format(result, self.possible))
+                raise mod_gpx.GPXException(f'Invalid value "{result}", possible: {self.possible}')
 
         return result
 
-    def to_xml(self, value, version, nsmap=None, prettyprint=True, indent=''):
+    def to_xml(self, value: Any, version: str, nsmap: Any=None, prettyprint: bool=True, indent: str='') -> Optional[str]:
         if value is None:
             return ''
         if not prettyprint:
             indent = ''
         if self.attribute:
-            return '{0}="{1}"'.format(self.attribute, mod_utils.make_str(value))
+            return '{}="{}"'.format(self.attribute, mod_utils.make_str(value))
         elif self.type_converter:
             value = self.type_converter.to_string(value)
-        return mod_utils.to_xml(self.tag, content=value, escape=True,
-                                prettyprint=prettyprint, indent=indent)
+        if self.tag:
+            return mod_utils.to_xml(self.tag, content=value, escape=True,
+                                    prettyprint=prettyprint, indent=indent)
+        return ''
 
 
 class GPXComplexField(AbstractGPXField):
-    def __init__(self, name, classs, tag=None, is_list=None):
+    def __init__(self, name: str, classs: Any, tag: Optional[str]=None, is_list: bool=False) -> None:
         AbstractGPXField.__init__(self, is_list=is_list)
         self.name = name
         self.tag = tag or name
         self.classs = classs
 
-    def from_xml(self, node, version):
+    def from_xml(self, node: Any, version: str) -> Any:
         if self.is_list:
             result = []
             for child in node:
@@ -241,7 +245,7 @@ class GPXComplexField(AbstractGPXField):
                 return None
             return gpx_fields_from_xml(self.classs, field_node, version)
 
-    def to_xml(self, value, version, nsmap=None, prettyprint=True, indent=''):
+    def to_xml(self, value: Any, version: str, nsmap: Dict[str, str]={}, prettyprint: bool=True, indent: str='') -> str:
         if not prettyprint:
             indent = ''
         if self.is_list:
@@ -261,12 +265,12 @@ class GPXEmailField(AbstractGPXField):
     """
     Converts GPX1.1 email tag group from/to string.
     """
-    def __init__(self, name, tag=None):
+    def __init__(self, name: str, tag: Optional[str]=None):
         AbstractGPXField.__init__(self, is_list=False)
         self.name = name
         self.tag = tag or name
 
-    def from_xml(self, node, version):
+    def from_xml(self, node: Any, version: str) -> Any:
         """
         Extract email address.
 
@@ -283,9 +287,9 @@ class GPXEmailField(AbstractGPXField):
 
         email_id = email_node.get('id')
         email_domain = email_node.get('domain')
-        return '{0}@{1}'.format(email_id, email_domain)
+        return f'{email_id}@{email_domain}'
 
-    def to_xml(self, value, version, nsmap=None, prettyprint=True, indent=''):
+    def to_xml(self, value: Any, version: str, nsmap: Optional[Dict[str, str]]=None, prettyprint: bool=True, indent: str='') -> str:
         """
         Write email address to XML
 
@@ -312,7 +316,7 @@ class GPXEmailField(AbstractGPXField):
             email_domain = 'unknown'
 
         return ('\n' + indent +
-                '<{0} id="{1}" domain="{2}" />'.format(self.tag,
+                '<{} id="{}" domain="{}" />'.format(self.tag,
                                                        email_id, email_domain))
 
 
@@ -320,12 +324,12 @@ class GPXExtensionsField(AbstractGPXField):
     """
     GPX1.1 extensions <extensions>...</extensions> key-value type.
     """
-    def __init__(self, name, tag=None, is_list=True):
+    def __init__(self, name: str, tag: Optional[str]=None, is_list: bool=True) -> None:
         AbstractGPXField.__init__(self, is_list=is_list)
         self.name = name
         self.tag = tag or 'extensions'
 
-    def from_xml(self, node, version):
+    def from_xml(self, node: Any, version: str) -> Any:
         """
         Build a list of extension Elements.
 
@@ -336,7 +340,7 @@ class GPXExtensionsField(AbstractGPXField):
         Returns:
             a list of Element objects
         """
-        result = []
+        result: Any = []
         extensions_node = node.find(self.tag)
         if extensions_node is None:
             return result
@@ -344,7 +348,7 @@ class GPXExtensionsField(AbstractGPXField):
             result.append(mod_copy.deepcopy(child))
         return result
 
-    def _resolve_prefix(self, qname, nsmap):
+    def _resolve_prefix(self, qname: str, nsmap: Dict[str, str]) -> str:
         """
         Convert a tag from Clark notation into prefix notation.
 
@@ -369,7 +373,7 @@ class GPXExtensionsField(AbstractGPXField):
                     break
         return qname
 
-    def _ETree_to_xml(self, node, nsmap=None, prettyprint=True, indent=''):
+    def _ETree_to_xml(self, node: Any, nsmap: Dict[str, str]={}, prettyprint: bool=True, indent: str='') -> str:
         """
         Serialize ETree element and all subelements.
 
@@ -397,7 +401,7 @@ class GPXExtensionsField(AbstractGPXField):
         result.append('\n' + indent + '<' + prefixedname)
         for attrib, value in node.attrib.items():
             attrib = self._resolve_prefix(attrib, nsmap)
-            result.append(' {0}="{1}"'.format(attrib, value))
+            result.append(f' {attrib}="{value}"')
         result.append('>')
         if node.text is not None:
              result.append(node.text.strip())
@@ -421,7 +425,7 @@ class GPXExtensionsField(AbstractGPXField):
 
         return ''.join(result)
 
-    def to_xml(self, value, version, nsmap=None, prettyprint=True, indent=''):
+    def to_xml(self, value: Any, version: str, nsmap: Dict[str, str]={}, prettyprint: bool=True, indent: str='') -> str:
         """
         Serialize list of ETree.
 
@@ -457,7 +461,7 @@ class GPXExtensionsField(AbstractGPXField):
 # Utility methods:
 # ----------------------------------------------------------------------------------------------------
 
-def _check_dependents(gpx_object, fieldname):
+def _check_dependents(gpx_object: Any, fieldname: str) -> Tuple[str, str]:
     """
     Check for data in subelements.
 
@@ -484,8 +488,8 @@ def _check_dependents(gpx_object, fieldname):
         return '/' + field, field # No child has data
     return '', fieldname # No children
 
-def gpx_fields_to_xml(instance, tag, version, custom_attributes=None,
-                      nsmap=None, prettyprint=True, indent=''):
+def gpx_fields_to_xml(instance: Any, tag: str, version: str, custom_attributes: Dict[str, str]={},
+                      nsmap: Dict[str, str]={}, prettyprint: bool=True, indent: str='') -> str:
     if not prettyprint:
         indent = ''
     fields = instance.gpx_10_fields
@@ -497,17 +501,17 @@ def gpx_fields_to_xml(instance, tag, version, custom_attributes=None,
     if tag:
         body.append('\n' + indent + '<' + tag)
         if tag == 'gpx':  # write nsmap in root node
-            body.append(' xmlns="{0}"'.format(nsmap['defaultns']))
+            body.append(' xmlns="{}"'.format(nsmap['defaultns']))
             namespaces = set(nsmap.keys())
             namespaces.remove('defaultns')
             for prefix in sorted(namespaces):
                 body.append(
-                    ' xmlns:{0}="{1}"'.format(prefix, nsmap[prefix])
+                    ' xmlns:{}="{}"'.format(prefix, nsmap[prefix])
                 )
         if custom_attributes:
             # Make sure to_xml() always return attributes in the same order:
             for key in sorted(custom_attributes.keys()):
-                body.append(' {0}="{1}"'.format(key, mod_utils.make_str(custom_attributes[key])))
+                body.append(' {}="{}"'.format(key, mod_utils.make_str(custom_attributes[key])))
     suppressuntil = ''
     for gpx_field in fields:
         # strings indicate non-data container tags with subelements
@@ -524,13 +528,13 @@ def gpx_fields_to_xml(instance, tag, version, custom_attributes=None,
                         body.append('>')
                         tag_open = False
                     if gpx_field[0] == '/':
-                        body.append('\n' + indent + '<{0}>'.format(gpx_field))
+                        body.append('\n' + indent + f'<{gpx_field}>')
                         if prettyprint and len(indent) > 1:
                             indent = indent[:-2]
                     else:
                         if prettyprint:
                             indent += '  '
-                        body.append('\n' + indent + '<{0}'.format(gpx_field))
+                        body.append('\n' + indent + f'<{gpx_field}')
                         tag_open = True
         elif not suppressuntil:
             value = getattr(instance, gpx_field.name)
@@ -556,7 +560,7 @@ def gpx_fields_to_xml(instance, tag, version, custom_attributes=None,
     return ''.join(body)
 
 
-def gpx_fields_from_xml(class_or_instance, node, version):
+def gpx_fields_from_xml(class_or_instance: Any, node: str, version: str) -> Any:
     if mod_inspect.isclass(class_or_instance):
         result = class_or_instance()
     else:
@@ -566,7 +570,7 @@ def gpx_fields_from_xml(class_or_instance, node, version):
     if version == '1.1':
         fields = result.gpx_11_fields
 
-    node_path = [node]
+    node_path: List[Union[str, int]] = [node]
 
     for gpx_field in fields:
         current_node = node_path[-1]
@@ -578,7 +582,7 @@ def gpx_fields_from_xml(class_or_instance, node, version):
                 if current_node is None:
                     node_path.append(None)
                 else:
-                    node_path.append(current_node.find(gpx_field))
+                    node_path.append(cast(str, current_node).find(gpx_field))
         else:
             if current_node is not None:
                 value = gpx_field.from_xml(current_node, version)
@@ -589,16 +593,12 @@ def gpx_fields_from_xml(class_or_instance, node, version):
 
     return result
 
-def gpx_check_slots_and_default_values(classs):
+def gpx_check_slots_and_default_values(classs: Callable[[], Any]) -> None:
     """
     Will fill the default values for this class. Instances will inherit those
     values so we don't need to fill default values for every instance.
-    This method will also fill the attribute gpx_field_names with a list of
-    gpx field names. This can be used
     """
-    fields = classs.gpx_10_fields + classs.gpx_11_fields
-
-    gpx_field_names = []
+    fields = classs.gpx_10_fields + classs.gpx_11_fields # type: ignore
 
     instance = classs()
 
@@ -607,34 +607,24 @@ def gpx_check_slots_and_default_values(classs):
         attributes = list(filter(lambda x : not callable(getattr(instance, x)), attributes))
         attributes = list(filter(lambda x : not x.startswith('gpx_'), attributes))
     except Exception as e:
-        raise Exception('Error reading attributes for %s: %s' % (classs.__name__, e))
+        raise Exception(f'Error reading attributes for {classs.__name__}: {e}')
 
     attributes.sort()
     slots = list(classs.__slots__)
     slots.sort()
 
     if attributes != slots:
-        raise Exception('Attributes for %s is\n%s but should be\n%s' % (classs.__name__, attributes, slots))
+        raise Exception(f'Attributes for {classs.__name__} is\n{attributes} but should be\n{slots}')
 
     for field in fields:
         if not isinstance(field, str):
+            value: Any = None
             if field.is_list:
                 value = []
-            else:
-                value = None
             try:
                 actual_value = getattr(instance, field.name)
             except:
-                raise Exception('%s has no attribute %s' % (classs.__name__, field.name))
-            if value != actual_value:
+                raise Exception(f'{classs.__name__} has no attribute {field.name}')
+            if field.name != "latitude" and field.name != "longitude" and value != actual_value:
                 raise Exception('Invalid default value %s.%s is %s but should be %s'
                                 % (classs.__name__, field.name, actual_value, value))
-            #print('%s.%s -> %s' % (classs, field.name, value))
-            if not field.name in gpx_field_names:
-                gpx_field_names.append(field.name)
-
-    gpx_field_names = tuple(gpx_field_names)
-##    if not hasattr(classs, '__slots__') or not classs.__slots__ or classs.__slots__ != gpx_field_names:
-##        try: slots = classs.__slots__
-##        except Exception as e: slots = '[Unknown:%s]' % e
-##        raise Exception('%s __slots__ invalid, found %s, but should be %s' % (classs, slots, gpx_field_names))
