@@ -17,7 +17,7 @@ import math as mod_math
 import xml.sax.saxutils as mod_saxutils
 import datetime as mod_datetime
 
-from typing import Any, AnyStr, List, Optional
+from typing import Any, AnyStr, List, Optional, cast
 
 def to_xml(tag: str, attributes: Any=None, content: Any=None, default: Any=None, escape: bool=False, prettyprint: bool=True, indent: str='') -> str:
     if not prettyprint:
@@ -71,7 +71,7 @@ def total_seconds(timedelta: mod_datetime.timedelta) -> float:
     """ Some versions of python don't have the timedelta.total_seconds() method. """
     if timedelta is None:
         return None
-    return (timedelta.days * 86400) + timedelta.seconds
+    return timedelta.total_seconds()
 
 
 def make_str(s: AnyStr) -> str:
